--Fortune Wheel Module, created by Itam, recoded by Hanakocz
--can be used by any map, just use Public.fortune_wheel(player) function to trigger it
--certain rolls are not included if some modules aren't used, for example RPG outcomes work only on maps with RPG.

local Global = require 'utils.global'
local Event = require 'utils.event'
local Public = {}

------global table for settings--------
local fortune_table = {}

Global.register(
    fortune_table,
    function(tbl)
        fortune_table = tbl
    end
)

function Public.reset_table()
  for k, _ in pairs(fortune_table) do
    fortune_table[k] = nil
  end
  fortune_table.coins_enabled = false
  fortune_table.coins_scale = 1
end

function Public.get_table()
    return fortune_table
end

local on_init = function()
    Public.reset_table()
end

Event.on_init(on_init)

------check if modules used-----
--dynamically enables possible outcomes if module is in use
local function rpg()
  if is_loaded('modules.rpg.main') then
    return {active = true, RPG = is_loaded('modules.rpg.table')}
  else
    return {active = false, RPG = nil}
  end
end

local function loot()
  if is_loaded('maps.amap.main') then
    return {active = true, Loot = is_loaded('maps.amap.loot')}
  else
    return {active = false, Loot = nil}
  end
end

----fortune wheel effects-------

local function lose_rpg_stats(player)
  if not rpg().active then return end
  local RPG = rpg().RPG
  if not RPG then return end
  local rpg_t = RPG.get('rpg_t')
  local stats = {
    [1] = {stat = "strength", message = {'amap.nopoint'}},
    [2] = {stat = "magicka", message = {'amap.nb16'}},
    [3] = {stat = "dexterity", message = {'amap.nb17'}},
    [4] = {stat = "vitality", message = {'amap.nb18'}}
  }
  local roll = math.random(1, #stats)
  if rpg_t[player.index][stats[roll].stat] < 15 then
    rpg_t[player.index][stats[roll].stat] = 10
  else
    rpg_t[player.index][stats[roll].stat] = rpg_t[player.index][stats[roll].stat] - 5
  end
  player.print(stats[roll].message)
end

local function gain_rpg_stats(player)
  if not rpg().active then return end
  local RPG = rpg().RPG
  if not RPG then return end
  local rpg_t = RPG.get('rpg_t')
  local stats = {
    [1] = {stat = "strength", message = {'amap.nb6'}},
    [2] = {stat = "magicka", message = {'amap.nb5'}},
    [3] = {stat = "dexterity", message = {'amap.nb4'}},
    [4] = {stat = "vitality", message = {'amap.nb3'}}
  }
  local roll = math.random(1, #stats)
  rpg_t[player.index][stats[roll].stat] = rpg_t[player.index][stats[roll].stat] + 15
  player.print(stats[roll].message)
end

local function gain_rpg_points(player)
  if not rpg().active then return end
  local RPG = rpg().RPG
  if not RPG then return end
  local rpg_t = RPG.get('rpg_t')
  rpg_t[player.index].points_to_distribute = rpg_t[player.index].points_to_distribute + 10
  player.print({'amap.nb2'})
end

local function rpg_xp(player)
  if not rpg().active then return end
  local RPG = rpg().RPG
  if not RPG then return end
  local rpg_t = RPG.get('rpg_t')
  local xp_values = {
    [1] = {value = -1000, message = {'amap.nb8'}},
    [2] = {value = 100, message = {'amap.nb12'}},
  }
  local roll = math.random(1, #xp_values)
  if rpg_t[player.index].xp + xp_values[roll].value < 0 then
    rpg_t[player.index].xp = 0
  else
    rpg_t[player.index].xp = rpg_t[player.index].xp + xp_values[roll].value
  end
  player.print(xp_values[roll].message)
end

local function gain_coins(player)
  local table = Public.get_table()
  local coef = table.coins_scale
  local coin_values = {
    [1] = {value = -1000 * coef, message = {'amap.sorry'}},
    [2] = {value = 1000 * coef, message = {'amap.nbone'}},
    [3] = {value = 10000 * coef, message = {'amap.nb13'}},
  }
  local roll = math.random(1, #coin_values)
  if coin_values[roll].value > 0 then
    player.insert{name='coin', count = coin_values[roll].value}
  else
    player.remove_item{name='coin', count = coin_values[roll].value}
  end
  player.print(coin_values[roll].message)
end

local function gain_items(player)
  local items = {
    [1] = {item = {name='raw-fish', count = 1}, message = {'amap.nb9'}},
    [2] = {item = {name='raw-fish', count = 100}, message = {'amap.nb10'}},
    [3] = {item = {name='distractor-capsule', count = 10}, message = {'amap.nb11'}},
  }
  local roll = math.random(1, #items)
  player.insert(items[roll].item)
  player.print(items[roll].message)
end

local function treasure(player)
  local Loot = loot().Loot
  if not Loot then return end
  Loot.cool(player.surface, player.surface.find_non_colliding_position("steel-chest", player.position, 20, 1, true) or player.position, 'steel-chest', 200)
  player.print({'amap.nb14'})
end

local function nothing(player)
  player.print({'amap.nb7'})
end

local function confusion(player)
  local messages = {
    [1] = {'amap.what'},
    [2] = "HAHAHAHAHAHAHAHA...what?",
    [3] = "?_?",
    [4] = {"amap.nb7"},
  }
  player.print(messages[math.random(1, #messages)])
end

----fortune wheel-----

local function fortune_list()
  local table = Public.get_table()
  local list = {
    [1] = {
      name = lose_rpg_stats,
      enabled = rpg().active,
      weight = 4
    },
    [2] = {
      name = gain_rpg_stats,
      enabled = rpg().active,
      weight = 4
    },
    [3] = {
      name = rpg_points,
      enabled = rpg().active,
      weight = 1
    },
    [4] = {
      name = rpg_xp,
      enabled = rpg().active,
      weight = 1
    },
    [5] = {
      name = gain_coins,
      enabled = table.coins_enabled,
      weight = 3
    },
    [6] = {
      name = gain_items,
      enabled = true,
      weight = 3
    },
    [7] = {
      name = treasure,
      enabled = loot().active,
      weight = 1
    },
    [8] = {
      name = nothing,
      enabled = true,
      weight = 1
    },
    [9] = {
      name = confusion,
      enabled = true,
      weight = 1
    }
  }
  return list
end

--magic function to roll random value from list, but value with weight 2 has twice as big chance to happen than value with weight 1.
local function raffle(values,weights) --arguments of the form {[a] = A, [b] = B, ...} and {[a] = a_weight, [b] = b_weight, ...} or just {a,b,c,...} and {1,2,3...}
	local total_weight = 0
	for k,w in pairs(weights) do
		assert(values[k])
		if w > 0 then
			total_weight = total_weight + w
		end
		-- negative weights treated as zero
	end
	assert(total_weight > 0)

	local cumulative_probability = 0
	local rng = math.random()
	for k,v in pairs(values) do
		assert(weights[k])
		cumulative_probability = cumulative_probability + (weights[k] / total_weight)
		if rng <= cumulative_probability then
			return v
		end
	end
end

--main function to use, checks the list and constructs the list of enabled options, then picks one randomly based on weights.
function Public.fortune_wheel(player)
  local indexes = {}
  local weights = {}
  local list = fortune_list()
  for index, list_item in pairs(list) do
    if list[index].enabled then
      indexes[#indexes + 1] = index
      weights[#weights + 1] = list[index].weight
    end
  end
  local roll = raffle(indexes, weights)
  if list[roll] then list[roll].name(player) end
end

return Public
